"""Database helpers for sentence persistence."""

from dataclasses import dataclass
from typing import Optional

import sqlalchemy

DB_NAME = "db.json"


class NoDatabaseSpecified(Exception):
    """No database connection was specified during init."""


class SentenceIdNotFound(Exception):
    """Could not find a previously generated sentence with specified ID."""


class DatabaseInsertFailed(Exception):
    """Raised when a sentence cannot be inserted into the database."""


@dataclass
class DatabaseConnection:
    """Stores info related to database connection."""

    engine: sqlalchemy.engine.Engine
    metadata: sqlalchemy.MetaData
    sentence_table: sqlalchemy.Table


def create_database_connection(
    voice_name: str,
    database: Optional[sqlalchemy.engine.Engine],
) -> Optional[DatabaseConnection]:
    if database is None:
        return None

    metadata = sqlalchemy.MetaData()
    sentence_table = sqlalchemy.Table(
        voice_name,
        metadata,
        sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True),
        sqlalchemy.Column("sentence", sqlalchemy.String, unique=True, nullable=False),
    )
    metadata.create_all(database)
    return DatabaseConnection(
        metadata=metadata, sentence_table=sentence_table, engine=database
    )


def insert_sentence_into_db(db: Optional[DatabaseConnection], sentence: str) -> int:
    if db is None:
        raise NoDatabaseSpecified

    insert_query = db.sentence_table.insert().values(sentence=sentence)
    with db.engine.begin() as conn:
        result = conn.execute(insert_query)

    if result.inserted_primary_key is None:
        raise DatabaseInsertFailed

    sentence_id = result.inserted_primary_key[0]
    if sentence_id is None:
        raise DatabaseInsertFailed

    return int(sentence_id)


def sentence_exists(db: Optional[DatabaseConnection], sentence: str) -> bool:
    if db is None:
        raise NoDatabaseSpecified

    select_query = db.sentence_table.select().where(
        db.sentence_table.c.sentence == sentence
    )
    with db.engine.connect() as conn:
        result = conn.execute(select_query)

    return bool(result.all())


def get_generated_sentences_list(db: Optional[DatabaseConnection]) -> list[str]:
    if db is None:
        raise NoDatabaseSpecified

    select_query = db.sentence_table.select()
    with db.engine.connect() as conn:
        result = conn.execute(select_query)

    return [row.sentence for row in result.all()]


def get_generated_sentences_dict(db: Optional[DatabaseConnection]) -> dict[int, str]:
    if db is None:
        raise NoDatabaseSpecified

    select_query = db.sentence_table.select()
    with db.engine.connect() as conn:
        result = conn.execute(select_query)

    return {row.id: row.sentence for row in result.all()}


def get_generated_sentence_from_id(
    db: Optional[DatabaseConnection], sentence_id: str
) -> Optional[str]:
    if db is None:
        raise NoDatabaseSpecified

    select_query = db.sentence_table.select().where(
        db.sentence_table.c.id == int(sentence_id)
    )
    with db.engine.connect() as conn:
        result = conn.execute(select_query)

    sentences = result.all()
    if not sentences:
        return None

    return str(sentences[0].sentence)



"""Manages multiple voices"""

import logging
import threading
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Optional, TypeVar, Union, Type
from types import TracebackType

import sqlalchemy

from hlvox.multi_voice import MultiVoice
from hlvox.single_voice import SingleVoice
from hlvox.voice import Voice

log = logging.getLogger(__name__)
T = TypeVar("T")


class DuplicateVoice(Exception):
    """Raised when duplicate voices are found"""


@dataclass
class RemoteDatabaseInfo:
    """Database connection information if using remote database (sql)"""

    name: str
    url: str
    port: int
    username: str
    password: str


@dataclass
class LocalDatabaseInfo:
    """Database info if using local file-based database (sqlite)"""

    base_path: Path


class Manager:
    """Manages multiple voices"""

    def __init__(
        self,
        voices_path: Union[Path, str],
        database_info: Optional[Union[LocalDatabaseInfo, RemoteDatabaseInfo]],
    ):
        self.voices_path = Path(voices_path)
        self._database_info = database_info
        self.voices: Dict[str, Union[MultiVoice, SingleVoice]] = {}
        self._lock = threading.RLock()

        self.scan_voices()

    def __enter__(self) -> "Manager":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.exit()

    def _create_db_path(self, databases_path: Path, voice_name: str) -> Path:
        db_path = databases_path / voice_name
        db_path.mkdir(parents=True, exist_ok=True)
        return db_path

    def _create_db(
        self,
        database_info: Union[LocalDatabaseInfo, RemoteDatabaseInfo],
        voice_name: str,
    ) -> sqlalchemy.engine.Engine:
        dbi = database_info
        if isinstance(dbi, LocalDatabaseInfo):
            path = self._create_db_path(
                databases_path=dbi.base_path, voice_name=voice_name
            )
            return sqlalchemy.create_engine(f"sqlite:///{path}/db.sqlite")
        return sqlalchemy.create_engine(
            f"postgresql+psycopg2://{dbi.username}:{dbi.password}@{dbi.url}:{dbi.port}/{dbi.name}",
            pool_pre_ping=True,
        )

    def scan_voices(self) -> None:
        """Scan for voices in the voices_path, updating the voices dictionary"""
        with self._lock:
            # Clean up existing voices and their database connections, if any
            for voice in self.voices.values():
                voice.exit()
            self.voices.clear()

            # Load new voices
            # This strangeness is to make mypy happy. There is probably a cleaner way to do it.
            single_voices = self._load_voices(self.voices_path)
            multi_voice = self._create_multi_voice(single_voices)
            self.voices.update(single_voices)
            self.voices["multi"] = multi_voice

    def _load_voices(self, path: Path) -> Dict[str, SingleVoice]:
        voices = {}
        voice_folders = list(x for x in path.iterdir() if x.is_dir())
        logging.info("Found %d voice folders", len(voice_folders))
        for voice_folder in voice_folders:
            voice_name = voice_folder.name.lower()
            logging.info("Loading voice %s", voice_name)
            if voice_name in voices:
                raise DuplicateVoice("Duplicate voice name found")

            database = None
            if self._database_info is not None:
                database = self._create_db(
                    database_info=self._database_info, voice_name=voice_name
                )
            new_voice = SingleVoice(
                name=voice_name, path=voice_folder, database=database
            )
            voices[new_voice.name] = new_voice
        logging.info("Loaded %d voices", len(voices))
        return voices

    def _create_multi_voice(self, voices: Dict[str, SingleVoice]) -> MultiVoice:
        database = None
        if self._database_info is not None:
            database = self._create_db(
                database_info=self._database_info, voice_name="multi"
            )
        return MultiVoice(
            voices=voices,
            database=database,
        )

    def get_voice_names(self) -> List[str]:
        """Gets names of available voices

        Returns:
            list -- list of voice name strings
        """

        with self._lock:
            voice_names = list(self.voices.keys())
            voice_names.sort()
            return voice_names

    def get_voice(self, name: str) -> Optional[Voice]:
        """Get voice of requested name

        Args:
            name ({string}): name of voice to get

        Returns:
            {voxvoice}: requested voice
        """
        with self._lock:
            if name in self.voices:
                return self.voices[name]
            return None

    def exit(self) -> None:
        """Exit all loaded voices"""
        with self._lock:
            for voice in self.voices.values():
                voice.exit()

    def run_with_voice(self, name: str, callback: Callable[[Voice], T]) -> Optional[T]:
        """Run callback with a voice while holding manager lock."""
        with self._lock:
            voice = self.voices.get(name)
            if voice is None:
                return None
            return callback(voice)



"""Defines core voice interfaces and compatibility exports."""

import logging
from types import TracebackType
from typing import Dict, List, Optional, Tuple, Type

import sqlalchemy

from .voice_audio import Audio
from .voice_db import (
    DB_NAME,
    DatabaseConnection,
    DatabaseInsertFailed,
    NoDatabaseSpecified,
    SentenceIdNotFound,
    create_database_connection,
    get_generated_sentence_from_id,
    get_generated_sentences_dict,
    get_generated_sentences_list,
    insert_sentence_into_db,
    sentence_exists,
)
from .voice_modifiers import (
    MODIFIERS,
    AcceleratorDirection,
    AcceleratorDirectionInvalid,
    AcceleratorModifier,
    CutModifier,
    Modifier,
    ModifierArgumentsInvalid,
    ModifierClass,
    ModifierFormatNotCorrect,
    Modifiers,
    NoOpModifier1,
    NoOpModifier2,
    PitchChangeModifier,
    SpeedChangeModifier,
    VolumeChangeModifier,
)
from .voice_sentence import (
    NoVoiceSpecified,
    PUNCTUATION_TIMING_SECONDS,
    Sentence,
    Word,
    extract_modifiers,
    process_sentence,
    split_sentence,
)

log = logging.getLogger(__name__)

__all__ = [
    "Audio",
    "DB_NAME",
    "DatabaseConnection",
    "DatabaseInsertFailed",
    "NoDatabaseSpecified",
    "SentenceIdNotFound",
    "MODIFIERS",
    "AcceleratorDirection",
    "AcceleratorDirectionInvalid",
    "AcceleratorModifier",
    "CutModifier",
    "Modifier",
    "ModifierArgumentsInvalid",
    "ModifierClass",
    "ModifierFormatNotCorrect",
    "Modifiers",
    "NoOpModifier1",
    "NoOpModifier2",
    "PitchChangeModifier",
    "SpeedChangeModifier",
    "VolumeChangeModifier",
    "NoVoiceSpecified",
    "PUNCTUATION_TIMING_SECONDS",
    "Sentence",
    "Word",
    "NoWordsFound",
    "DuplicateWords",
    "InconsistentAudioFormats",
    "NoAudioFormatFound",
    "FailedToSplit",
    "ModifierSyntaxError",
    "Voice",
]


class NoWordsFound(Exception):
    """Raised when a voice has no words."""


class DuplicateWords(Exception):
    """Raised when a voice has duplicate words."""


class InconsistentAudioFormats(Exception):
    """Raised when words have inconsistent audio formats."""


class NoAudioFormatFound(Exception):
    """Raised when no audio format can be found."""


class FailedToSplit(Exception):
    """Raised when a sentence cannot be split."""


class ModifierSyntaxError(Exception):
    """Raised when there is a problem with the modifier syntax."""


class Voice:
    """Base class for Voice-like interfaces.
    Intended to involve generation of audio files from some source (files, web, etc).
    """

    def __init__(
        self,
        name: str,
        database: Optional[sqlalchemy.engine.Engine],
    ) -> None:
        self.name = name
        self._db = create_database_connection(voice_name=self.name, database=database)
        self.words: List[Word] = []
        self.categories: Dict[str, List[str]] = {}

    def __enter__(self) -> "Voice":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.exit()

    def exit(self) -> None:
        """Clean up and close voice."""
        if self._db is not None:
            self._db.engine.dispose()

    def _insert_sentence_into_db(self, sentence: str) -> int:
        return insert_sentence_into_db(db=self._db, sentence=sentence)

    def _sentence_exists(self, sentence: str) -> bool:
        return sentence_exists(db=self._db, sentence=sentence)

    def _get_generated_sentences_list(self) -> list[str]:
        return get_generated_sentences_list(db=self._db)

    def _get_generated_sentences_dict(self) -> dict[int, str]:
        return get_generated_sentences_dict(db=self._db)

    def _get_generated_sentence_from_id(self, sentence_id: str) -> Optional[str]:
        return get_generated_sentence_from_id(db=self._db, sentence_id=sentence_id)

    def _get_sentence_info(self, words: List[Word]) -> Sentence:
        sayable_words, unsayable_words = self.get_sayable_unsayable(words)
        sayable_sent_arr = self._get_sayable_sentence_arr(words, sayable_words)
        sayable_sent_str = self._create_sentence_string(sayable_sent_arr)

        return Sentence(
            sentence=sayable_sent_str,
            sayable=sayable_words,
            unsayable=unsayable_words,
            sayable_sentence=sayable_sent_arr,
            audio=None,
            id=None,
        )

    def generate_audio_from_array(
        self, words: List[Word], dry_run: bool = False, save_to_db: bool = True
    ) -> Sentence:
        """Generates audio segment from sentence array."""
        sentence_info = self._get_sentence_info(words=words)

        if dry_run:
            return sentence_info

        log.debug("Generating %s", sentence_info.sentence)

        if len(sentence_info.sayable) == 0:
            log.warning(
                "Can't say any words in %s, not generating", sentence_info.sentence
            )
            return sentence_info

        if self._db and save_to_db and len(words) > 1:
            if not self._sentence_exists(sentence=sentence_info.sentence):
                sentence_id = self._insert_sentence_into_db(
                    sentence=sentence_info.sentence
                )
                sentence_info.id = str(sentence_id)

        words_audio = self._create_audio_segments(sentence_info.sayable_sentence)
        sentence_info.audio = self.assemble_audio_segments(words_audio)

        return sentence_info

    def _create_audio_segments(
        self,
        word_array: List[Word],  # pylint: disable=unused-argument
    ) -> List[Audio]:
        return []

    def generate_audio(self, sentence: str, dry_run: bool = False) -> Sentence:
        """Generates audio from the given sentence."""
        log.info("Asked to generate %s", sentence)
        split = self._split_sentence(sentence)
        processed_sentence = self.process_sentence(split, voice=self.name)
        return self.generate_audio_from_array(words=processed_sentence, dry_run=dry_run)

    def generate_audio_from_id(
        self, sentence_id: str, dry_run: bool = False
    ) -> Sentence:
        """Generate audio from an existing sentence, fetched by its ID."""
        log.info("Fetching audio for ID %s", sentence_id)
        sentence = self._get_generated_sentence_from_id(sentence_id)
        if not sentence:
            raise SentenceIdNotFound
        return self.generate_audio(sentence=sentence, dry_run=dry_run)

    @staticmethod
    def _split_sentence(sentence: str) -> List[Word]:
        return split_sentence(sentence)

    @staticmethod
    def _extract_modifiers(words: List[Word]) -> List[Word]:
        return extract_modifiers(words)

    @staticmethod
    def process_sentence(split_sent: List[Word], voice: Optional[str]) -> List[Word]:
        return process_sentence(split_sent=split_sent, voice=voice)

    def get_sayable_unsayable(self, words: List[Word]) -> Tuple[List[Word], List[Word]]:
        """Get words that are sayable or unsayable from a list of words."""
        sayable_words_set = set(self.words)
        sayable_words_set.update(
            [
                Word(word=punct, voice=self.name, is_punctuation=True)
                for punct in PUNCTUATION_TIMING_SECONDS
            ]
        )

        no_modifier_words = [word.without_modifiers() for word in words]
        words_set = set((dict.fromkeys(no_modifier_words)))

        unsayable_set = words_set - sayable_words_set
        sayable_set = words_set - unsayable_set
        unsayable = list(unsayable_set)
        unsayable.sort()
        sayable = list(sayable_set)
        sayable.sort()
        return sayable, unsayable

    def _get_sayable_sentence_arr(
        self, words: List[Word], sayable_words: List[Word]
    ) -> List[Word]:
        return [
            word
            for word in words
            if Word(
                word=word.word, voice=word.voice, is_punctuation=word.is_punctuation
            )
            in sayable_words
        ]

    def _create_sentence_string(self, words: List[Word]) -> str:
        if len(words) == 1:
            return words[0].as_str(with_voice=False)
        return " ".join([word.as_str(with_voice=False) for word in words])

    def get_generated_sentences(self) -> List[str]:
        return self._get_generated_sentences_list()

    def get_generated_sentences_dict(self) -> Dict[int, str]:
        return self._get_generated_sentences_dict()

    @staticmethod
    def assemble_audio_segments(segments: List[Audio]) -> Audio:
        """Assemble audio segments into one audio segment."""
        frame_rates = [word.get_sample_rate() for word in segments]
        frame_rate = min(frame_rates)

        sentence_audio = segments.pop(0)
        if sentence_audio.get_sample_rate() != frame_rate:
            sentence_audio.set_sample_rate(frame_rate)

        for word_audio in segments:
            if word_audio.get_sample_rate() != frame_rate:
                word_audio.set_sample_rate(frame_rate)
            sentence_audio = sentence_audio + word_audio

        return sentence_audio



"""Audio primitives for voice generation."""

from pathlib import Path
from typing import BinaryIO

import librosa
import numpy as np
import soundfile as sf  # type: ignore[import-untyped]


class Audio:
    """Represents an audio file."""

    def __init__(self, data: np.ndarray, sample_rate: float):
        self.data = data
        self.sample_rate = sample_rate

    @classmethod
    def from_file(cls, file_path: Path) -> "Audio":
        data, sample_rate = librosa.load(file_path, mono=True)
        return Audio(data=data, sample_rate=sample_rate)

    @classmethod
    def create_silence(cls, duration: float, sample_rate: float) -> "Audio":
        data = np.zeros(int(duration * sample_rate))
        return Audio(data=data, sample_rate=sample_rate)

    def get_sample_rate(self) -> float:
        return self.sample_rate

    def set_sample_rate(self, sample_rate: float) -> None:
        resampled_data = librosa.resample(
            y=self.data,
            orig_sr=self.sample_rate,
            target_sr=sample_rate,
        )
        self.data = resampled_data
        self.sample_rate = sample_rate

    def get_length_seconds(self) -> float:
        return librosa.get_duration(y=self.data, sr=self.sample_rate)

    def export(self, output: BinaryIO, audio_format: str) -> None:
        sf.write(
            file=output,
            data=self.data,
            samplerate=self.sample_rate,
            format=audio_format,
        )

    def __add__(self, other: "Audio") -> "Audio":
        if self.get_sample_rate() > other.get_sample_rate():
            other.set_sample_rate(self.get_sample_rate())
        elif self.get_sample_rate() < other.get_sample_rate():
            self.set_sample_rate(other.get_sample_rate())

        combined_data: np.ndarray = np.concatenate((self.data, other.data))
        return Audio(data=combined_data, sample_rate=self.get_sample_rate())



"""Modifier definitions and parsing for voice generation."""

import copy
import enum
from dataclasses import dataclass
from typing import Dict, List, Protocol, Type, TypeAlias, Union

import librosa

from .voice_audio import Audio


class ModifierFormatNotCorrect(Exception):
    """Modifiers not provided in the correct format."""


class AcceleratorDirectionInvalid(Exception):
    """Invalid accelerator direction."""


class ModifierArgumentsInvalid(Exception):
    """Invalid modifier arguments and/or format."""


class ModifierClass(Protocol):
    IDENTIFIER: str

    @classmethod
    def from_str(cls, string: str) -> "Modifier": ...


@dataclass
class Modifier:
    """Base class for sound modifiers."""

    IDENTIFIER = ""

    @classmethod
    def from_str(cls, string: str) -> "Modifier":  # pylint: disable=unused-argument
        return Modifier()

    def as_str(self) -> str:
        return ""

    def modify_audio(self, audio: Audio) -> Audio:
        return audio

    def __eq__(self, other: object) -> bool:
        raise NotImplementedError


@dataclass
class SpeedChangeModifier(Modifier):
    """Modify speed of audio without changing pitch."""

    IDENTIFIER = "s"

    def __init__(self, speed_parameter: float) -> None:
        if speed_parameter <= 0.0:
            raise ModifierArgumentsInvalid
        self.speed_parameter = float(speed_parameter)

    @classmethod
    def from_str(cls, string: str) -> "SpeedChangeModifier":
        speed_parameter_str = string.strip(SpeedChangeModifier.IDENTIFIER)
        return SpeedChangeModifier(speed_parameter=float(speed_parameter_str))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.speed_parameter}"

    def modify_audio(self, audio: Audio) -> Audio:
        stretched_data = librosa.effects.time_stretch(
            y=audio.data, rate=self.speed_parameter
        )
        return Audio(data=stretched_data, sample_rate=audio.sample_rate)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SpeedChangeModifier):
            return False
        return self.speed_parameter == other.speed_parameter


@dataclass
class PitchChangeModifier(Modifier):
    """Modify pitch of audio without changing speed."""

    IDENTIFIER = "p"

    def __init__(self, pitch_parameter: float) -> None:
        self.pitch_parameter = float(pitch_parameter)

    @classmethod
    def from_str(cls, string: str) -> "PitchChangeModifier":
        pitch_parameter_str = string.strip(PitchChangeModifier.IDENTIFIER)
        return PitchChangeModifier(pitch_parameter=float(pitch_parameter_str))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.pitch_parameter}"

    def modify_audio(self, audio: Audio) -> Audio:
        pitch_shifted_data = librosa.effects.pitch_shift(
            y=audio.data,
            sr=audio.sample_rate,
            n_steps=self.pitch_parameter,
        )
        return Audio(data=pitch_shifted_data, sample_rate=audio.sample_rate)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PitchChangeModifier):
            return False
        return self.pitch_parameter == other.pitch_parameter


class AcceleratorDirection(enum.Enum):
    """Direction for accelerator modifier."""

    FORWARD = enum.auto()
    REVERSE = enum.auto()

    @classmethod
    def from_str(cls, value: str) -> "AcceleratorDirection":
        if value == "f":
            return AcceleratorDirection.FORWARD
        if value == "r":
            return AcceleratorDirection.REVERSE
        raise AcceleratorDirectionInvalid

    def to_str(self) -> str:
        if self == AcceleratorDirection.FORWARD:
            return "f"
        if self == AcceleratorDirection.REVERSE:
            return "r"
        raise AcceleratorDirectionInvalid


@dataclass
class AcceleratorModifier(Modifier):
    """Accelerate a single word with modifiable pitch and speed curves."""

    IDENTIFIER = "a"

    def __init__(
        self, speed_increment: float, count: int, direction: AcceleratorDirection
    ) -> None:
        if speed_increment <= 0.0:
            raise ModifierArgumentsInvalid
        if count <= 1.0:
            raise ModifierArgumentsInvalid

        self.speed_increment = speed_increment
        self.count = int(count)
        self.direction = direction

    @classmethod
    def from_str(cls, string: str) -> "AcceleratorModifier":
        parameter_str = string.strip(cls.IDENTIFIER)
        components = parameter_str.split("+")
        if len(components) < 2 or len(components) > 3:
            raise ModifierFormatNotCorrect

        speed_increment = float(components[0])
        count = int(components[1])
        direction = AcceleratorDirection.FORWARD
        if len(components) == 3:
            direction = AcceleratorDirection.from_str(components[2])

        return AcceleratorModifier(
            speed_increment=speed_increment, count=count, direction=direction
        )

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.speed_increment}+{self.count}+{self.direction.to_str()}"

    def modify_audio(self, audio: Audio) -> Audio:
        segment: Audio | None = None
        increments: Union[range, reversed]

        if self.direction == AcceleratorDirection.REVERSE:
            increments = reversed(range(0, self.count))
        else:
            increments = range(0, self.count)

        for increment in increments:
            if increment == 0.0:
                speed_increment = 1.0
            else:
                speed_increment = increment * self.speed_increment

            speed_modifier = SpeedChangeModifier(speed_parameter=speed_increment)
            if speed_modifier != 1.0:
                modified_audio = speed_modifier.modify_audio(audio=audio)
            else:
                modified_audio = copy.deepcopy(audio)

            if segment is None:
                segment = modified_audio
            else:
                segment += modified_audio

        if segment is None:
            raise ModifierArgumentsInvalid("No segments were created")

        return Audio(data=segment.data, sample_rate=segment.sample_rate)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AcceleratorModifier):
            return False
        return (
            self.speed_increment == other.speed_increment
            and self.count == other.count
            and self.direction == other.direction
        )


@dataclass
class VolumeChangeModifier(Modifier):
    """Modify volume of audio."""

    IDENTIFIER = "v"

    def __init__(self, volume_parameter: float) -> None:
        self.volume_parameter = float(volume_parameter)

    @classmethod
    def from_str(cls, string: str) -> "VolumeChangeModifier":
        volume_parameter_str = string.strip(VolumeChangeModifier.IDENTIFIER)
        return VolumeChangeModifier(volume_parameter=float(volume_parameter_str))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.volume_parameter}"

    def modify_audio(self, audio: Audio) -> Audio:
        return Audio(
            data=audio.data * self.volume_parameter, sample_rate=audio.sample_rate
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, VolumeChangeModifier):
            return False
        return self.volume_parameter == other.volume_parameter


class CutModifier(Modifier):
    """Cut start and end of audio."""

    IDENTIFIER = "c"

    def __init__(self, start_time_s: float | None, end_time_s: float | None) -> None:
        if start_time_s is None and end_time_s is None:
            raise ModifierArgumentsInvalid
        if start_time_s is not None and start_time_s < 0:
            raise ModifierArgumentsInvalid
        if end_time_s is not None and end_time_s < 0:
            raise ModifierArgumentsInvalid
        if (
            start_time_s is not None
            and end_time_s is not None
            and start_time_s >= end_time_s
        ):
            raise ModifierArgumentsInvalid

        self.start_time_s = start_time_s
        self.end_time_s = end_time_s

    @classmethod
    def from_str(cls, string: str) -> "CutModifier":
        start_time_s_str, end_time_s_str = string.strip(CutModifier.IDENTIFIER).split(
            "-"
        )
        start_time_s = None if start_time_s_str == "" else float(start_time_s_str)
        end_time_s = None if end_time_s_str == "" else float(end_time_s_str)
        return CutModifier(start_time_s=start_time_s, end_time_s=end_time_s)

    def as_str(self) -> str:
        start_time_s_str = "" if self.start_time_s is None else f"{self.start_time_s}"
        end_time_s_str = "" if self.end_time_s is None else f"{self.end_time_s}"
        return f"{self.IDENTIFIER}{start_time_s_str}-{end_time_s_str}"

    def modify_audio(self, audio: Audio) -> Audio:
        start_index = (
            0
            if self.start_time_s is None
            else int(self.start_time_s * audio.sample_rate)
        )
        end_index = (
            len(audio.data)
            if self.end_time_s is None
            else int(self.end_time_s * audio.sample_rate)
        )
        return Audio(
            data=audio.data[start_index:end_index], sample_rate=audio.sample_rate
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CutModifier):
            return False
        return (
            self.start_time_s == other.start_time_s
            and self.end_time_s == other.end_time_s
        )


@dataclass
class NoOpModifier1(Modifier):
    """No-op modifier for testing."""

    IDENTIFIER = "z"

    def __init__(self, parameter: float) -> None:
        self.parameter = float(parameter)

    @classmethod
    def from_str(cls, string: str) -> "NoOpModifier1":
        parameter = string.strip(NoOpModifier1.IDENTIFIER)
        return NoOpModifier1(parameter=float(parameter))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.parameter}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoOpModifier1):
            return False
        return self.parameter == other.parameter


@dataclass
class NoOpModifier2(Modifier):
    """No-op modifier for testing."""

    IDENTIFIER = "x"

    def __init__(self, parameter: float) -> None:
        self.parameter = float(parameter)

    @classmethod
    def from_str(cls, string: str) -> "NoOpModifier2":
        parameter = string.strip(NoOpModifier2.IDENTIFIER)
        return NoOpModifier2(parameter=float(parameter))

    def as_str(self) -> str:
        return f"{self.IDENTIFIER}{self.parameter}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoOpModifier2):
            return False
        return self.parameter == other.parameter


MODIFIERS: Dict[str, Type[ModifierClass]] = {
    CutModifier.IDENTIFIER: CutModifier,
    VolumeChangeModifier.IDENTIFIER: VolumeChangeModifier,
    SpeedChangeModifier.IDENTIFIER: SpeedChangeModifier,
    PitchChangeModifier.IDENTIFIER: PitchChangeModifier,
    AcceleratorModifier.IDENTIFIER: AcceleratorModifier,
    NoOpModifier1.IDENTIFIER: NoOpModifier1,
    NoOpModifier2.IDENTIFIER: NoOpModifier2,
}

Modifiers: TypeAlias = List[Modifier]



"""Sentence and word parsing helpers for voice generation."""

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional

from .voice_audio import Audio
from .voice_modifiers import MODIFIERS, Modifier, Modifiers

log = logging.getLogger(__name__)

# How much delay should be added in place of punctuation (in milliseconds)
PUNCTUATION_TIMING_SECONDS = {
    ",": 0.250,
    ".": 0.500,
}


class NoVoiceSpecified(Exception):
    """Raised when no voice is specified."""


@dataclass
class Sentence:
    """Represents a sentence and its parts."""

    sentence: str
    sayable: List["Word"]
    unsayable: List["Word"]
    sayable_sentence: List["Word"]
    audio: Optional[Audio] = None
    id: Optional[str] = None


class Word:
    """Represents a word (or punctuation)."""

    def __init__(
        self,
        word: str,
        voice: Optional[str] = None,
        is_punctuation: bool = False,
        modifiers: Optional[Modifiers] = None,
    ) -> None:
        self.word = word
        self.voice = voice
        self.modifiers = modifiers if modifiers else []
        self.modifiers.sort(key=lambda modifier: modifier.IDENTIFIER)
        self.is_punctuation = is_punctuation

    def as_str(self, with_voice: bool = False) -> str:
        voice_string = f"{self.voice}:" if (with_voice and self.voice) else ""
        modifiers_string = ""
        if self.modifiers:
            modifier_strings = ",".join(
                [modifier.as_str() for modifier in self.modifiers]
            )
            modifiers_string = f"|{modifier_strings}"
        return f"{voice_string}{self.word}{modifiers_string}"

    def __str__(self) -> str:
        return self.as_str(with_voice=True)

    def __repr__(self) -> str:
        return self.as_str(with_voice=True)

    def without_modifiers(self) -> "Word":
        return Word(
            word=self.word, voice=self.voice, is_punctuation=self.is_punctuation
        )

    def __lt__(self, other: "Word") -> bool:
        return self.word < other.word

    def __gt__(self, other: "Word") -> bool:
        return self.word > other.word

    def __le__(self, other: "Word") -> bool:
        return self.word <= other.word

    def __ge__(self, other: "Word") -> bool:
        return self.word >= other.word

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Word):
            return False
        return (
            self.word == other.word
            and self.modifiers == other.modifiers
            and self.is_punctuation == other.is_punctuation
            and self.voice == other.voice
        )

    def __hash__(self) -> int:
        return hash(repr(self))


def split_sentence(sentence: str) -> List[Word]:
    return [Word(word=word) for word in sentence.lower().rstrip().split(" ")]


def _parse_modifiers(word: str) -> tuple[str, list[Modifier]]:
    if "|" not in word:
        return word, []

    word_string, _, modifiers_string = word.rpartition("|")
    modifier_strings = modifiers_string.split(",")

    # dict keeps insertion order and dedupes by identifier
    modifiers: Dict[str, Modifier] = {}
    for modifier_string in modifier_strings:
        if not modifier_string:
            continue
        modifier_class = MODIFIERS.get(modifier_string[0])
        if modifier_class is None:
            continue
        modifier = modifier_class.from_str(modifier_string)
        if modifier.IDENTIFIER in modifiers:
            continue
        modifiers[modifier.IDENTIFIER] = modifier

    return word_string, list(modifiers.values())


def extract_modifiers(words: List[Word]) -> List[Word]:
    """Extract modifiers from raw words without punctuation processing."""

    processed_words: list[Word] = []
    for word in words:
        word_string, modifiers = _parse_modifiers(word.word)
        processed_words.append(
            Word(
                word=word_string,
                voice=word.voice,
                is_punctuation=word.is_punctuation,
                modifiers=modifiers,
            )
        )
    return processed_words


def process_sentence(split_sent: List[Word], voice: Optional[str]) -> List[Word]:
    """Break sentence into word/punctuation elements and extract modifiers in one pass."""

    log.info("Processing sentence '%s'", split_sent)
    reduced_sent: list[Word] = []

    for item in split_sent:
        word_string, modifiers = _parse_modifiers(item.word)

        first_punct: Optional[str] = None
        try:
            first_punct = next(
                punct for punct in PUNCTUATION_TIMING_SECONDS if punct in word_string
            )
        except StopIteration:
            pass

        if first_punct:
            first_punct_ind = word_string.find(first_punct)

            if first_punct_ind >= 2 and word_string[first_punct_ind - 1] == ":":
                reduced_sent.append(
                    Word(
                        word=word_string[: first_punct_ind + 1],
                        voice=voice,
                        modifiers=modifiers,
                    )
                )
                if len(word_string) >= first_punct_ind:
                    first_punct_ind += 1
            else:
                if word_string[:first_punct_ind]:
                    reduced_sent.append(
                        Word(
                            word=word_string[:first_punct_ind],
                            voice=voice,
                            modifiers=modifiers,
                        )
                    )

            for punct in word_string[first_punct_ind:]:
                if punct in PUNCTUATION_TIMING_SECONDS:
                    reduced_sent.append(
                        Word(word=punct, voice=voice, is_punctuation=True)
                    )
        else:
            reduced_sent.append(
                Word(word=word_string, voice=voice, modifiers=modifiers)
            )

    reduced_sent = [value for value in reduced_sent if value.word != ""]
    log.info("Sentence processed: '%s'", reduced_sent)
    return reduced_sent
